import pkg_resources
from threading import Thread
from gi.repository import Gtk
from gi.repository import GObject
from mokaplayer.core.database import Song
from mokaplayer.core.fetchers import lyrics


class LyricsWindow():
    GLADE_FILE = pkg_resources.resource_filename('mokaplayer',
                                                 'data/ui/lyrics_window.ui')


    def __init__(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(self.GLADE_FILE)
        self.builder.connect_signals(self)
        self.window = self.builder.get_object('lyrics_window')
        self.lbl_title = self.builder.get_object('lbl_title')
        self.txt_lyrics = self.builder.get_object('txt_lyrics')

    def start_fetch(self, song):
        self.song = song
        self.txt_lyrics.get_buffer().set_text('')

        if self.song is not None:
            self.lbl_title.set_text(f'{self.song.Title} - {self.song.Artist}')
            Thread(target=self.fetch).start()
        else:
            self.lbl_title.set_text('')

    def fetch(self):
        ly = self.song.Lyrics

        if not ly:
            success, ly = lyrics.get(self.song.Title, self.song.Artist, self.song.Album)
            if success:
                self.song.Lyrics = ly
                self.song.save()
        else:
            ly = ly + ' (cached)'

        GObject.idle_add(lambda: self.on_fetch_finished(ly))

    def on_fetch_finished(self, ly):
        self.txt_lyrics.get_buffer().set_text(ly)

    def get_window(self):
        return self.window

    def on_btn_refresh_clicked(self, widget):
        if self.song is not None:
            self.song.Lyrics = None
            self.start_fetch(self.song)



import logging
import urwid

from gi.repository import GObject

from mokaplayer.core.helpers import time_helper


class MainScreen():
    def __init__(self, appconfig, userconfig, player):
        self.logger = logging.getLogger('MainWindow')
        self.appconfig = appconfig
        self.userconfig = userconfig
        self.player = player

        self.player.state_changed.subscribe(self.on_player_state_changed)
        self.player.audio_changed.subscribe(self.on_audio_changed)
        self.player.volume_changed.subscribe(self.on_volume_changed)

        self.__set_current_song_info()
        self.on_volume_changed()

        GObject.timeout_add(750, self.on_tick, None)
        txt = urwid.Text(u"Hello World")
        fill = urwid.Filler(txt, 'top')
        loop = urwid.MainLoop(fill)
        loop.run()


        self.logger.info('Screen loaded')

    def __set_current_song_info(self):
        self.logger.debug("Setting currrent song info")

    def show(self):
        pass

    def on_audio_changed(self):
        GObject.idle_add(self.__set_current_song_info)

    def on_player_state_changed(self):
        GObject.idle_add(self.__set_current_song_info)

    def on_volume_changed(self):
        pass

    def on_tick(self, data):
        position = self.player.streamer.position
        duration = self.player.streamer.duration
        fraction = position / duration if duration else 0
        position_text = time_helper.seconds_to_string(position)
        duration_text = time_helper.seconds_to_string(duration)

        # self.lbl_current_time.set_text(f'{position_text} / {duration_text}')
        # self.prb_current_time.set_fraction(fraction)

        return True



import logging


class KeyboardClient:
    def __init__(self, player):
        self.player = player
        self.logger = logging.getLogger('KeyboardClient')

        try:
            import PyHook3
            hm = PyHook3.HookManager()
            hm.KeyDown = self._on_pyhook_event
            hm.HookKeyboard()
            return
        except ImportError:
            self.logger.warning('PyHook3 is needed on Window for MediaKey binding')

    def _on_XF86AudioNext(self, key):
        self.logger.info(key)
        self.player.next()

    def _on_XF86AudioPrev(self, key):
        self.logger.info(key)
        self.player.prev()

    def _on_XF86AudioPlay(self, key):
        self.logger.info(key)
        if self.player.streamer.state == self.player.streamer.State.PAUSED:
            self.player.play()
        elif self.player.streamer.state == self.player.streamer.State.PLAYING:
            self.player.pause()

    def _on_pyhook_event(self, event):
        if event.Key == 'Media_Next_Track':
            self._on_XF86AudioNext('XF86AudioNext')
        elif event.Key == 'Media_Prev_Track':
            self._on_XF86AudioPrev('XF86AudioPrev')
        elif event.Key == 'Media_Play_Pause':
            self._on_XF86AudioPlay('XF86AudioPlay')
        return True



import peewee
from mokaplayer.core.database import database_context


class Album(peewee.Model):
    AlbumId = peewee.PrimaryKeyField()
    Name = peewee.CharField(index=True)
    Path = peewee.CharField(index=True)
    Year = peewee.CharField(null=True)
    Cover = peewee.CharField(null=True)
    Description = peewee.CharField(null=True)
    Artist = peewee.CharField(null=True)

    class Meta:
        database = database_context



from .provider_ultimateguitar import ProviderUltimateGuitar


def search(title, artist):
    return ProviderUltimateGuitar.search(title, artist)


def fetch_ascii_tab(url):
    return ProviderUltimateGuitar.fetch_ascii_tab(url)


def download_guitar_pro_tab(url, directory):
    return ProviderUltimateGuitar.download_guitar_pro_tab(url, directory)



from .provider_base import ProviderBase


class ProviderDarkLyrics(ProviderBase):
    def get_name(self):
        return "DarkLyrics"

    def get_url(self, title, artist, album):
        return 'http://www.darklyrics.com/lyrics/{}/{}.html'.format(self.encode(artist),
                                                                    self.encode(album))

    def get_xpath(self, title, artist, album):
        return '//div[@class="lyrics"]//text()'

    def get_regex(self, title, artist, album):
        return f'\d+\.\s{title.lower()}([\S\s]*?)(\d+\.\s|$)'

    def get_node_separator(self):
        return ''

    def encode(self, param):
        param = (
            param
            .lower()
            .replace(' ', '')
            .replace('(', '')
            .replace(')', '')
            .replace('.', '')
            .replace('[', '')
            .replace(']', '')
            .replace('/', '')
            .replace('?', '')
            .replace('!', '')
            .replace("'", '')
        )

        return super().encode(param)



import collections
import os
import pathlib
import shutil
import logging
import yaml
import appdirs
import pkg_resources


def merge(d1, d2):
    for k in d2:
        if k in d1 and isinstance(d1[k], dict) and isinstance(d2[k], dict):
            merge(d1[k], d2[k])
        else:
            d1[k] = d2[k]


class UserConfig(object):
    """Read, parse and manage the user configuration file
    """

    APPLICATION_DEFAULT_CONFIG_PATH = pkg_resources.resource_filename('mokaplayer',
                                                                      'data/default_user_config.yaml')

    def __init__(self, path=None):
        """ Create the user configuration file if not found
            Parse the user configuration file
        """
        self._path = path
        self._data = {}
        if path is not None:
            if self.get_file().exists():
                self.read()
            elif path:
                self.create()

    def read(self):
        """ Parse the YAML file
        """
        try:
            with open(self.APPLICATION_DEFAULT_CONFIG_PATH, 'rt') as stream:
                self._data = yaml.load(stream)
            with open(self.get_file(), 'rt') as stream:
                merge(self._data, yaml.load(stream))
        except:
            logging.critical('Could not read the user configuration file', exc_info=True)

    def save(self):
        """ Save the YAML file
        """
        try:
            if self._path is not None:
                with open(self.get_file(), 'wt') as stream:
                    yaml.dump(self._data, stream, default_flow_style=False)
        except:
            logging.critical('Could not save the user configuration file', exc_info=True)

    def get_file(self):
        """ Return a file object of the current user configuration file
        """
        return pathlib.Path(self._path)

    def create(self):
        """ Copy the default user configuration file, then parse it
        """
        shutil.copy(self.APPLICATION_DEFAULT_CONFIG_PATH, self.get_file())
        self.read()

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value



import threading
import time
import datetime
import arrow
import logging
import pkg_resources

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from mokaplayer.core.player import Player
from mokaplayer.core.helpers import time_helper
from mokaplayer.ui.gtk.helper import image_helper, file_helper
from mokaplayer.ui.gtk.adapter import AdapterSong
from mokaplayer.ui.gtk.windows import AboutWindow
from mokaplayer.ui.gtk.windows import LyricsWindow
from mokaplayer.ui.gtk.windows import TabsWindow
from mokaplayer.ui.gtk.windows import TagsEditorWindow
from mokaplayer.ui.gtk.windows import HelpShortcutsWindow


class MainWindow(Gtk.Window):
    """ Main window
    """

    GLADE_FILE = pkg_resources.resource_filename('mokaplayer',
                                                 'data/ui/main_window.ui')

    def __init__(self, appconfig, userconfig, player):
        super().__init__(title="MokaPlayer", default_width=1366, default_height=768)

        self.logger = logging.getLogger('MainWindow')
        self.appconfig = appconfig
        self.userconfig = userconfig
        self.player = player
        self.set_icon_from_file(pkg_resources.resource_filename('mokaplayer', 'data/mokaplayer.png'))

        if self.userconfig['gtk']['darktheme']:
            settings = Gtk.Settings.get_default()
            settings.set_property("gtk-application-prefer-dark-theme", True)

        self.connect("destroy", self.on_window_destroy)
        self.connect("key-press-event", self.on_window_key_press)

        self.builder = Gtk.Builder()
        self.builder.add_from_file(self.GLADE_FILE)
        self.__get_object()
        self.__init_sort_radio()
        self.__init_gridview_columns()

        self.builder.connect_signals(self)
        self.player.state_changed.subscribe(self.on_player_state_changed)
        self.player.audio_changed.subscribe(self.on_audio_changed)
        self.player.volume_changed.subscribe(self.on_volume_changed)

        self.__set_model(AdapterSong.create_store())
        self.__set_current_song_info()
        self.on_volume_changed()

        threading.Thread(target=self.__create_model).start()

        GObject.timeout_add(750, self.on_tick, None)
        self.logger.info('Window loaded')

        if not self.player.library.is_musics_folder_valid():
            self.__ask_for_music_folder()

    def __ask_for_music_folder(self):
        dialog = Gtk.FileChooserDialog("Select where is your music", self,
                                       Gtk.FileChooserAction.SELECT_FOLDER,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        "Select", Gtk.ResponseType.OK))

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self.player.library.musics_folder = dialog.get_filename()
            self.on_library_scan_activate(None)
        dialog.destroy()

    def __init_gridview_columns(self):
        columns = self.gridview.get_columns()
        font = self.userconfig['grid']['font']
        for col in columns:
            col.set_visible(False)

            if font:
                for cell in col.get_cells():
                    cell.set_property('font', font)

        for name in reversed(self.userconfig['grid']['columns']):
            for col in columns:
                if name == col.get_title():
                    col.set_visible(True)
                    self.gridview.move_column_after(col, None)

    def __create_model(self):
        self.logger.info('Creating ListStore')
        start = time.perf_counter()
        model = AdapterSong.create_store()

        order = self.userconfig['grid']['order']['field']
        desc = self.userconfig['grid']['order']['desc']

        data = self.player.library.search_song(order, desc)

        for row in data:
            model.insert_with_valuesv(-1, AdapterSong.create_col_number(),
                                      AdapterSong.create_row(row))

        end = time.perf_counter()
        self.logger.info('ListStore created in {:.3f} seconds'.format(end - start))

        GObject.idle_add(lambda: self.__set_model(model))

    def __set_model(self, model):
        self.model = model.filter_new()
        self.model.set_visible_func(self.__model_filter_func)
        self.gridview.set_model(self.model)

    def __get_object(self):
        self.content = self.builder.get_object('content')
        self.gridview = self.builder.get_object('gridview')
        self.col_added = self.builder.get_object('col_added')
        self.col_length = self.builder.get_object('col_length')
        self.lbl_current_title = self.builder.get_object('lbl_current_title')
        self.lbl_current_song_infos = self.builder.get_object('lbl_current_song_infos')
        self.volume_scale = self.builder.get_object('volume_scale')
        self.btn_previous = self.builder.get_object('btn_previous')
        self.btn_play = self.builder.get_object('btn_play')
        self.btn_next = self.builder.get_object('btn_next')
        self.img_current_album = self.builder.get_object('img_current_album')
        self.img_play = self.builder.get_object('img_play')
        self.img_pause = self.builder.get_object('img_pause')
        self.txt_search = self.builder.get_object('txt_search')
        self.lbl_current_time = self.builder.get_object('lbl_current_time')
        self.prb_current_time = self.builder.get_object('prb_current_time')
        self.radio_sort_artist = self.builder.get_object('radio_sort_artist')
        self.radio_sort_album = self.builder.get_object('radio_sort_album')
        self.radio_sort_title = self.builder.get_object('radio_sort_title')
        self.radio_sort_year = self.builder.get_object('radio_sort_year')
        self.radio_sort_length = self.builder.get_object('radio_sort_length')
        self.radio_sort_added = self.builder.get_object('radio_sort_added')
        self.radio_sort_played = self.builder.get_object('radio_sort_played')
        self.chk_sort_desc = self.builder.get_object('chk_sort_desc')
        self.menu_gridview = self.builder.get_object('menu_gridview')
        self.search_bar = self.builder.get_object('search_bar')
        self.add(self.content)

    def __set_current_song_info(self):
        self.logger.debug("Setting currrent song info")
        image_size = self.userconfig['header']['image_size']
        line1 = self.userconfig['header']['line1']
        line2 = self.userconfig['header']['line2']

        song = self.player.library.get_song(self.player.queue.peek())
        album = None

        if song:
            album = self.player.library.get_album(song.Album, song.AlbumArtist)
            self.lbl_current_title.set_text(line1.format(**song._data))
            self.lbl_current_song_infos.set_text(line2.format(**song._data))
        else:
            self.lbl_current_title.set_text('')
            self.lbl_current_song_infos.set_text('')

        self.img_current_album.set_from_pixbuf(image_helper.load(album.Cover if album else None,
                                                                 image_size, image_size))

    def __set_current_play_icon(self):
        if self.player.streamer.state == self.player.streamer.State.PLAYING:
            self.btn_play.set_image(self.img_pause)
        else:
            self.btn_play.set_image(self.img_play)

    def __model_filter_func(self, model, iter, data):
        return AdapterSong.search(self.txt_search.get_text(), model[iter])

    def __focus_song(self, path):
        for row in self.model:
            if row[0] == path:
                self.gridview.get_selection().select_path(row.path)
                self.gridview.set_cursor(row.path)
                self.gridview.scroll_to_cell(row.path, use_align=True)
                self.gridview.grab_focus()
                break

    def __show_lyrics(self, path):
        song = self.player.library.get_song(path)
        w = LyricsWindow()
        w.start_fetch(song)
        w.get_window().set_transient_for(self)
        w.get_window().show()

    def __show_tabs(self, path):
        song = self.player.library.get_song(path)
        w = TabsWindow()
        w.start_fetch(song)
        w.get_window().set_transient_for(self)
        w.get_window().show()

    def __show_tagseditor(self, paths):
        songs = self.player.library.get_songs(paths)
        w = TagsEditorWindow(songs)
        w.get_window().set_transient_for(self)
        w.get_window().show()

    def __library_scan(self):
        self.player.library.sync()
        GObject.idle_add(self.on_library_scan_finished)

    def __library_artwork(self):
        self.player.library.sync_artwork()
        GObject.idle_add(self.on_library_artworks_finished)

    def on_window_destroy(self, widget):
        self.player.save()
        Gtk.main_quit()

    def on_window_key_press(self, widget, event):
        keyval_name = Gdk.keyval_name(event.keyval)
        ctrl = (event.state & Gdk.ModifierType.CONTROL_MASK)

        if ctrl and keyval_name == 'f':
            search_mode = self.search_bar.get_search_mode()
            self.search_bar.set_search_mode(not search_mode)
            if not search_mode:
                self.txt_search.grab_focus()
        elif ctrl and keyval_name == 'l':
            self.__show_lyrics(self.player.queue.peek())
        elif ctrl and keyval_name == 'o':
            self.__focus_song(self.player.queue.peek())
        elif ctrl and keyval_name == 'Left':
            self.player.streamer.volume -= 0.05
        elif ctrl and keyval_name == 'Right':
            self.player.streamer.volume += 0.05
        elif keyval_name == "Escape":
            self.search_bar.set_search_mode(False)
        elif keyval_name == "space" and not self.txt_search.has_focus():
            self.player.toggle()
        else:
            return False

        return True

    def on_gridview_button_press_event(self, sender, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3:
            self.menu_gridview.popup(None, None, None, None, event.button, event.time)
            return True

    def on_menu_gridview_append_activate(self, widget):
        self.player.queue.append(self.__get_selected_songs_in_gridview())

    def on_menu_gridview_insert_activate(self, widget):
        self.player.queue.prepend(self.__get_selected_songs_in_gridview())

    def on_menu_gridview_replace_activate(self, widget):
        self.player.queue.clear()
        self.player.queue.append(self.__get_selected_songs_in_gridview())

    def on_menu_gridview_edit_activate(self, widget):
        selected_songs = self.__get_selected_songs_in_gridview()
        if any(selected_songs):
            self.__show_tagseditor(self.__get_selected_songs_in_gridview())

    def on_menu_gridview_tabs_activate(self, widget):
        selected_songs = self.__get_selected_songs_in_gridview()
        if any(selected_songs):
            self.__show_tabs(selected_songs[0])

    def on_menu_gridview_lyrics_activate(self, widget):
        selected_songs = self.__get_selected_songs_in_gridview()
        if any(selected_songs):
            self.__show_lyrics(selected_songs[0])

    def on_menu_gridview_folder_activate(self, widget):
        selected_songs = self.__get_selected_songs_in_gridview()
        if any(selected_songs):
            file_helper.open_folder(selected_songs[0])

    def __get_selected_songs_in_gridview(self):
        model, pathlist = self.gridview.get_selection().get_selected_rows()
        result = []
        for path in pathlist:
            result.append(model[path][0])
        return result

    def on_gridview_row_activated(self, widget, path, column):
        self.player.seek(self.model[path][0])

    def on_btn_next_clicked(self, widget):
        self.player.next()

    def on_btn_play_clicked(self, widget):
        self.player.toggle()
        self.__set_current_play_icon()

    def on_btn_previous_clicked(self, widget):
        self.player.prev()

    def on_audio_changed(self):
        GObject.idle_add(self.__set_current_song_info)
        GObject.idle_add(self.__set_current_play_icon)

    def on_player_state_changed(self):
        GObject.idle_add(self.__set_current_play_icon)

    def on_volume_changed(self):
        self.volume_scale.set_value(self.player.streamer.volume * 100)

    def on_volume_scale_value_changed(self, widget):
        self.player.streamer.volume = self.volume_scale.get_value() / 100

    def on_txt_search_key_press_event(self, widget, event):
        keyval_name = Gdk.keyval_name(event.keyval)
        if keyval_name == "Return" and any(self.model):
            self.__focus_song(self.model[0][0])

    def on_txt_search_search_changed(self, widget):
        self.model.refilter()

    def on_player_open_stream_activate(self, event):
        pass

    def on_player_preferences_activate(self, event):
        pass

    def on_player_quit_activate(self, event):
        Gtk.main_quit()

    def on_library_scan_activate(self, event):
        threading.Thread(target=self.__library_scan).start()

    def on_library_scan_finished(self):
        self.on_sort_radio_toggled(None)

    def on_library_artworks_activate(self, event):
        threading.Thread(target=lambda: self.__library_artwork()).start()

    def on_library_artworks_finished(self):
        self.__set_current_song_info()

    def on_queue_add_activate(self, event):
        self.player.queue.clear()
        self.player.queue.append([x.Path for x in self.player.library.search_song()])
        self.player.queue.seek(self.player.queue.pop())

    def on_queue_clear_activate(self, event):
        self.player.queue.clear()

    def on_queue_shuffle_activate(self, event):
        self.player.queue.shuffle()

    def on_help_shortcuts_activate(self, event):
        shortcuts_window = HelpShortcutsWindow.get_diaglog()
        shortcuts_window.set_transient_for(self)
        shortcuts_window.show()

    def on_help_about_activate(self, event):
        about_window = AboutWindow.get_diaglog()
        about_window.set_transient_for(self)
        about_window.show()

    def on_prp_current_time_click(self, widget, event):
        width = self.prb_current_time.get_allocated_width()
        duration = self.player.streamer.duration
        self.player.streamer.position = duration * event.x / width

    def on_tick(self, data):
        position = self.player.streamer.position
        duration = self.player.streamer.duration
        fraction = position / duration if duration else 0
        position_text = time_helper.seconds_to_string(position)
        duration_text = time_helper.seconds_to_string(duration)

        self.lbl_current_time.set_text(f'{position_text} / {duration_text}')
        self.prb_current_time.set_fraction(fraction)

        return True

    def __init_sort_radio(self):
        self.chk_sort_desc.set_active(self.userconfig['grid']['order']['desc'])
        if self.userconfig['grid']['order']['field'] == 'Artist':
            self.radio_sort_artist.set_active(True)
        elif self.userconfig['grid']['order']['field'] == 'Album':
            self.radio_sort_album.set_active(True)
        elif self.userconfig['grid']['order']['field'] == 'Title':
            self.radio_sort_title.set_active(True)
        elif self.userconfig['grid']['order']['field'] == 'Length':
            self.radio_sort_length.set_active(True)
        elif self.userconfig['grid']['order']['field'] == 'Year':
            self.radio_sort_year.set_active(True)
        elif self.userconfig['grid']['order']['field'] == 'Added':
            self.radio_sort_added.set_active(True)
        elif self.userconfig['grid']['order']['field'] == 'Played':
            self.radio_sort_played.set_active(True)

    def on_sort_radio_toggled(self, widget):
        if widget is None or widget.get_active() or not isinstance(widget, Gtk.RadioButton):
            self.userconfig['grid']['order']['field'] = ''
            self.userconfig['grid']['order']['desc'] = self.chk_sort_desc.get_active()

            if self.radio_sort_artist.get_active():
                self.userconfig['grid']['order']['field'] = 'Artist'
            elif self.radio_sort_album.get_active():
                self.userconfig['grid']['order']['field'] = 'Album'
            elif self.radio_sort_title.get_active():
                self.userconfig['grid']['order']['field'] = 'Title'
            elif self.radio_sort_length.get_active():
                self.userconfig['grid']['order']['field'] = 'Length'
            elif self.radio_sort_year.get_active():
                self.userconfig['grid']['order']['field'] = 'Year'
            elif self.radio_sort_added.get_active():
                self.userconfig['grid']['order']['field'] = 'Added'
            elif self.radio_sort_played.get_active():
                self.userconfig['grid']['order']['field'] = 'Played'

            threading.Thread(target=self.userconfig.save).start()
            threading.Thread(target=self.__create_model).start()



from .provider_base import ProviderBase


class ProviderGenius(ProviderBase):
    def get_name(self):
        return "Genius"

    def get_url(self, title, artist, album):
        return 'https://genius.com/{}-{}-lyrics'.format(self.encode(artist),
                                                        self.encode(title))

    def get_xpath(self, title, artist, album):
        return '//div[@class="lyrics"]//text()'

    def get_regex(self, title, artist, album):
        return ''

    def get_node_separator(self):
        return ''

    def encode(self, param):
        param = (
            param
            .lower()
            .replace(' ', '-')
            .replace('(', '')
            .replace(')', '')
            .replace('.', '')
            .replace('[', '')
            .replace(']', '')
            .replace('/', '')
            .replace('?', '')
            .replace('!', '')
            .replace("'", '')
        )

        return super().encode(param)



from gi.repository import Gtk


class AdapterTab:
    @staticmethod
    def create_row(tab):
        return [
            str(tab['type']),
            tab['name'],
            tab['rating'],
            tab['url']
        ]

    @staticmethod
    def create_store():
        return Gtk.ListStore(str, str, int, str)

    @staticmethod
    def create_col_number():
        return [0, 1, 2, 3]



import datetime
import time
import pathlib
import logging
import json
import gzip
from threading import Thread

from mokaplayer.core.queue import Queue
from mokaplayer.core.event import Event
from mokaplayer.core.library import Library
from mokaplayer.core.streamer import Streamer
from mokaplayer.core.keyboard import KeyboardClient


class Player(object):

    def __init__(self, appconfig, userconfig):
        self.logger = logging.getLogger("Player")
        self.appconfig = appconfig
        self.userconfig = userconfig
        self.state_changed = Event()
        self.audio_changed = Event()
        self.volume_changed = Event()
        self.queue = Queue()
        self.streamer = Streamer(about_to_finish=self.__about_to_finish,
                                 audio_changed=self.__audio_changed,
                                 notify_volume=self.__notify_volume)

        self.library = Library(self.appconfig, self.userconfig)
        KeyboardClient(self)

    def play(self):
        self.logger.debug("Play")
        if self.streamer.state == Streamer.State.PAUSED:
            self.streamer.resume()
            self.state_changed.fire()
        elif self.streamer.state == Streamer.State.STOPED and self.queue.peek():
            self.streamer.play(self.queue.peek())
            self.state_changed.fire()

    def seek(self, path):
        self.logger.debug("Seek " + str(path))
        self.__set_play_count()
        self.queue.seek(path)
        self.streamer.play(path)

    def stop(self):
        self.logger.debug("Stop")
        self.__set_play_count()
        self.streamer.stop()
        self.queue.pop()
        self.state_changed.fire()

    def pause(self):
        self.logger.debug("Pause")
        self.streamer.pause()
        self.state_changed.fire()

    def toggle(self):
        self.logger.debug("Toggle")
        if self.streamer.state == Streamer.State.PLAYING:
            self.streamer.pause()
            self.state_changed.fire()
        elif self.streamer.state == Streamer.State.PAUSED:
            self.streamer.resume()
            self.state_changed.fire()
        elif self.streamer.state == Streamer.State.STOPED and self.queue.peek():
            self.streamer.play(self.queue.peek())
            self.state_changed.fire()

    def next(self):
        self.logger.debug("Next")
        if len(self.queue):
            self.__set_play_count()
            self.queue.next()
            self.streamer.play(self.queue.peek())
        else:
            self.stop()

        self.state_changed.fire()

    def prev(self):
        self.logger.debug("Prev")
        if len(self.queue):
            self.__set_play_count()
            self.queue.prev()
            self.streamer.play(self.queue.peek())
        else:
            self.stop()

        self.state_changed.fire()

    def restore(self):
        try:
            self.logger.info("Restoring")
            if pathlib.Path(self.appconfig.PLAYER_CACHE_FILE).is_file():
                with gzip.open(self.appconfig.PLAYER_CACHE_FILE, 'rt') as file:
                    data = json.load(file)

                    self.streamer.volume = data['Volume']
                    self.queue.clear()
                    self.queue.append(data['Queue'])
                    self.seek(data['Current'])
                    self.pause()
                    time.sleep(0.3)
                    self.streamer.position = data['Position']
        except Exception as e:
            self.logger.exception('Could not restore player state')

    def save(self):
        try:
            self.logger.info("Saving")
            pathlib.Path(self.appconfig.PLAYER_CACHE_FILE).parent.mkdir(exist_ok=True)
            with gzip.open(self.appconfig.PLAYER_CACHE_FILE, 'wt') as file:
                json.dump({
                    "Queue": list(self.queue),
                    "Current": self.queue.peek(),
                    "Position": self.streamer.position,
                    "Volume": self.streamer.volume,
                }, file)
        except Exception as e:
            self.logger.exception('Could not save player state')

    def __about_to_finish(self, data):
        self.logger.debug("About to finish")
        self.__set_play_count()
        self.queue.next()
        self.streamer.stream = self.queue.peek()

    def __audio_changed(self):
        self.logger.debug("Audio changed")
        self.audio_changed.fire()

    def __notify_volume(self, data, data2):
        self.volume_changed.fire()

    def __set_play_count(self):
        song = self.library.get_song(self.queue.peek())
        try:
            percent = self.streamer.position / self.streamer.duration
        except ZeroDivisionError:
            percent = 0

        if song is not None and percent > 0.85:
            song.Played += 1
            song.Last_played = datetime.datetime.now()
            Thread(target=song.save).start()



import pkg_resources
import ast
import logging
import arrow

from gi.repository import Gtk
from gi.repository import GObject

from mokaplayer.core.database import database_context
from mokaplayer.core.helpers import time_helper


class TagsEditorWindow:

    GLADE_FILE = pkg_resources.resource_filename('mokaplayer',
                                                 'data/ui/tagseditor_window.ui')

    def __init__(self, songs):
        self.songs = songs

        self.builder = Gtk.Builder()
        self.builder.add_from_file(self.GLADE_FILE)
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("tagseditor_window")
        self.txt_Title = self.builder.get_object("txt_Title")
        self.txt_Artist = self.builder.get_object("txt_Artist")
        self.txt_ArtistAlbum = self.builder.get_object("txt_ArtistAlbum")
        self.txt_Album = self.builder.get_object("txt_Album")
        self.txt_TrackNumber = self.builder.get_object("txt_TrackNumber")
        self.txt_Year = self.builder.get_object("txt_Year")
        self.txt_Genre = self.builder.get_object("txt_Genre")
        self.txt_Added = self.builder.get_object("txt_Added")
        self.txt_Comment = self.builder.get_object("txt_Comment")
        self.txt_Length = self.builder.get_object("txt_Length")
        self.txt_Channels = self.builder.get_object("txt_Channels")
        self.txt_Bitrate = self.builder.get_object("txt_Bitrate")
        self.txt_Path = self.builder.get_object("txt_Path")

        self.load_text()

    def load_text(self):
        self.set_text(self.txt_Path, lambda s: s.Path)
        self.set_text(self.txt_Bitrate, lambda s: s.Bitrate)
        self.set_text(self.txt_Channels, lambda s: s.Channels)
        self.set_text(self.txt_Length,
                      lambda s: time_helper.seconds_to_string(s.Length))
        self.set_text(self.txt_Title, lambda s: s.Title)
        self.set_text(self.txt_Artist, lambda s: s.Artist)
        self.set_text(self.txt_Album, lambda s: s.Album)
        self.set_text(self.txt_Genre, lambda s: s.Genre)
        self.set_text(self.txt_Year, lambda s: s.Year)
        self.set_text(self.txt_Comment, lambda s: s.Comment)
        self.set_text(self.txt_Added, lambda s: str(s.Added))
        self.set_text(self.txt_ArtistAlbum, lambda s: s.AlbumArtist)
        self.set_text(self.txt_TrackNumber, lambda s: s.Tracknumber)

    def get_window(self):
        return self.window

    def set_text(self, textentry, field):
        list_value = [field(s) for s in self.songs]
        # If every item are the same http://stackoverflow.com/a/3844948
        if list_value.count(list_value[0]) == len(list_value):
            textentry.set_text(str(list_value[0]))
        else:
            textentry.set_text(str(list_value))

    def get_value(self, textentry):
        text = textentry.get_text()

        if text.lstrip().startswith('[') and text.rstrip().endswith(']'):
            return ast.literal_eval(textentry.get_text())
        else:
            return [text for x in range(len(self.songs))]

    def on_btnSave_clicked(self, button):
        try:
            titles = self.get_value(self.txt_Title)
            artists = self.get_value(self.txt_Artist)
            artistalbums = self.get_value(self.txt_ArtistAlbum)
            albums = self.get_value(self.txt_Album)
            tracknumbers = self.get_value(self.txt_TrackNumber)
            years = self.get_value(self.txt_Year)
            addeds = self.get_value(self.txt_Added)
            genres = self.get_value(self.txt_Genre)
            comments = self.get_value(self.txt_Comment)

            with database_context.atomic():
                for index, song in enumerate(self.songs):
                    song.Title = titles[index]
                    song.Artist = artists[index]
                    song.AlbumArtist = artistalbums[index]
                    song.Album = albums[index]
                    song.Tracknumber = tracknumbers[index]
                    song.Year = years[index]
                    song.Comment = comments[index]
                    song.Genre = genres[index]
                    song.Added = arrow.get(addeds[index]).naive
                    song.save()
                    song.write_tags()

            self.window.destroy()

        except Exception as e:
            logging.exception("Could not save tags")

    def on_btnCancel_clicked(self, button):
        self.window.destroy()



import datetime
import logging
import peewee
import taglib

from mokaplayer.core.database import database_context


class Song(peewee.Model):
    songId = peewee.PrimaryKeyField()
    Path = peewee.CharField(index=True)
    Title = peewee.CharField(null=True, index=True)
    Album = peewee.CharField(null=True, index=True)
    Artist = peewee.CharField(null=True, index=True)
    AlbumArtist = peewee.CharField(null=True, index=True)
    Tracknumber = peewee.IntegerField(null=True, index=True)
    Genre = peewee.CharField(null=True)
    Discnumber = peewee.IntegerField(null=True, index=True)
    Comment = peewee.CharField(null=True)
    Year = peewee.CharField(null=True, index=True)
    Label = peewee.CharField(null=True)
    Lyrics = peewee.CharField(null=True)
    Added = peewee.DateField(default=datetime.date.today, index=True)
    Last_played = peewee.DateTimeField(null=True)
    Played = peewee.IntegerField(default=0)
    Length = peewee.IntegerField(null=True)
    Channels = peewee.IntegerField(null=True)
    Bitrate = peewee.IntegerField(null=True)

    class Meta:
        database = database_context

    def __get_tag_value(self, tags, name):
        l = tags.get(name, [''])
        return l[0] if any(l) else ''

    def write_tags(self):
        """ Write tags in the file
        """
        try:
            fileref = taglib.File(self.Path)

            fileref.tags['TITLE'] = [self.Title]
            fileref.tags['ALBUM'] = [self.Album]
            fileref.tags['ARTIST'] = [self.Artist]
            fileref.tags['DISCNUMBER'] = [str(self.Discnumber)]
            fileref.tags['COMMENT'] = [self.Comment]
            fileref.tags['DATE'] = [self.Year]
            fileref.tags['LABEL'] = [self.Label]
            fileref.tags['GENRE'] = [self.Genre]
            fileref.tags['TRACKNUMBER'] = [str(self.Tracknumber)]
            fileref.tags['ALBUMARTIST'] = [self.AlbumArtist]

            fileref.save()
        except:
            logging.exception('Could not write tag to ' + self.Path)

    def read_tags(self):
        """ Read tags from the file
        """
        try:
            fileref = taglib.File(self.Path)
            self.Title = self.__get_tag_value(fileref.tags, 'TITLE')
            self.Album = self.__get_tag_value(fileref.tags, 'ALBUM')
            self.Artist = self.__get_tag_value(fileref.tags, 'ARTIST')
            self.Discnumber = self.__get_tag_value(fileref.tags, 'DISCNUMBER')
            self.Comment = self.__get_tag_value(fileref.tags, 'COMMENT')
            self.Year = self.__get_tag_value(fileref.tags, 'DATE')[:4]
            self.Label = self.__get_tag_value(fileref.tags, 'LABEL')
            self.Lyrics = self.__get_tag_value(fileref.tags, 'LYRICS')
            self.Genre = self.__get_tag_value(fileref.tags, 'GENRE')
            self.Tracknumber = self.__get_tag_value(fileref.tags, 'TRACKNUMBER')
            self.AlbumArtist = self.__get_tag_value(fileref.tags, 'ALBUMARTIST')
            self.Length = fileref.length
            self.Channels = fileref.channels
            self.Bitrate = fileref.bitrate

            try:
                self.Tracknumber = int(self.Tracknumber.split('/')[0])
            except ValueError:
                self.Tracknumber = 0

            try:
                self.Discnumber = int(self.Discnumber.split('/')[0])
            except ValueError:
                self.Discnumber = 0

            self.AlbumArtist = self.AlbumArtist or self.Artist

        except:
            logging.exception('Could not read tag from ' + self.Path)



import pathlib
import pkg_resources
from gi.repository import GdkPixbuf


def load(filename, width, height):
    if not filename or not pathlib.Path(filename).is_file():
        filename = pkg_resources.resource_filename('mokaplayer', 'data/placeholder.png')

    pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
    pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
    return pixbuf



from .provider_base import ProviderBase


class ProviderMetrolyrics(ProviderBase):
    def get_name(self):
        return "metrolyrics"

    def get_url(self, title, artist, album):
        return 'http://www.metrolyrics.com/{}-lyrics-{}.html'.format(self.encode(title),
                                                                     self.encode(artist))

    def get_xpath(self, title, artist, album):
        return '//div[contains(@class, "js-lyric-text")]/p//text()'

    def get_regex(self, title, artist, album):
        return ''

    def get_node_separator(self):
        return ''

    def encode(self, param):
        param = (
            param
            .lower()
            .replace(' ', '-')
            .replace('(', '')
            .replace(')', '')
            .replace('.', '')
            .replace('/', '')
            .replace("'", '')
            .replace(',', '')
        )

        return super().encode(param)



import glob
import os
import re

from setuptools import setup, find_packages

def get_readme():
    with open('PYPI_README.rst') as f:
        return f.read()

def get_version():
    with open('mokaplayer/__init__.py') as f:
        version_match = re.search(r"__version__ = '(.*)'", f.read())
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")

def get_data_files():
    yield ('share/applications', ['mokaplayer/data/mokaplayer.desktop'])

    for path in glob.glob('mokaplayer/data/icons/hicolor/*x*'):
        folder = os.path.basename(path)
        yield (f'share/icons/hicolor/{folder}/apps', [f'{path}/apps/mokaplayer.png'])

setup(
    name="MokaPlayer",
    license="MIT",
    version=get_version(),
    author="Vincent Bédard",
    description="A simple music player (GTK)",
    long_description=get_readme(),
    keywords='mokaplayer music player tags tabs lyrics musicplayer gapless gtk',
    url="https://github.com/vedard/MokaPlayer/",
    packages=find_packages(exclude=["test"]),
    zip_safe=False,
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        'pyyaml',
        'appdirs',
        'peewee',
        'requests',
        'lxml',
        'arrow',
        'appdirs',
        'pytaglib'
    ],
    data_files= list(get_data_files()),
    entry_points={
        'console_scripts': [
            'mokaplayer=mokaplayer:main',
        ],
    },
)



from gi.repository import Gtk
import arrow
import re

from mokaplayer.core.helpers import time_helper


class AdapterSong:
    @staticmethod
    def create_row(song):
        return [
            song.Path,
            song.Title or 'Unkown',
            song.AlbumArtist or 'Unkown',
            song.Album or 'Unkown',
            time_helper.seconds_to_string(song.Length),
            song.Year or '0000',
            arrow.Arrow(song.Added.year, song.Added.month, song.Added.day).humanize(),
            song.Played,
            song.Genre,
            str(song.Tracknumber)
        ]

    @staticmethod
    def create_store():
        return Gtk.ListStore(str, str, str, str, str, str, str, str, str, str)

    @staticmethod
    def create_col_number():
        return [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    @staticmethod
    def search(text, row):
        if not text:
            return True
        else:
            try:
                return any(re.search(text, col, re.RegexFlag.IGNORECASE) for col in row)
            except re.error:
                return False  # regex syntax error



import gi
gi.require_version('Gtk', '3.0')

import pkg_resources
import argparse
import signal
import logging
import threading

from gi.repository import Gtk
from gi.repository import Gio

from mokaplayer.ui.gtk.windows.main_window import MainWindow
from mokaplayer.ui.ncurse.screens.main_screen import MainScreen
from mokaplayer.core.player import Player
from mokaplayer.config import appconfig as AppConfig
from mokaplayer.config.userconfig import UserConfig


class Application():
    def __init__(self):
        self.init_args()
        self.init_configs()
        self.init_logging()
        self.init_player()
        self.init_signals()
        self.init_window()

    def run(self):
        Gtk.main()

    def init_window(self):
        if self.args.ncurses:
            self.screen = MainScreen(self.appconfig, self.userconfig, self.player)
            self.screen.show()
        else:
            self.window = MainWindow(self.appconfig, self.userconfig, self.player)
            self.window.show_all()

    def init_args(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--debug', help='Load the developpement configuration', required=False, action='store_true')
        parser.add_argument('--ncurses', help='Terminal User Interface', required=False, action='store_true')
        self.args = parser.parse_args()

    def init_configs(self):
        self.appconfig = AppConfig.Developpement if self.args.debug else AppConfig.Production
        self.userconfig = UserConfig(self.appconfig.USER_CONFIG_FILE)

    def init_logging(self):
        logging.basicConfig(level=getattr(logging, self.appconfig.LOG_LEVEL))
        logging.getLogger("peewee").setLevel(logging.INFO)

    def init_player(self):
        self.player = Player(self.appconfig, self.userconfig)
        threading.Thread(target=self.player.restore).start()

    def init_signals(self):

        def handler(signum, frame):
            self.player.save()
            Gtk.main_quit()

        # Some singal are not availble on every OS
        if hasattr(signal, 'SIGTERM'):
            signal.signal(signal.SIGTERM, handler)
        if hasattr(signal, 'SIGINT'):
            signal.signal(signal.SIGINT, handler)
        if hasattr(signal, 'SIGHUP'):
            signal.signal(signal.SIGHUP, handler)
        if hasattr(signal, 'SIGQUIT'):
            signal.signal(signal.SIGQUIT, handler)